from django.apps import AppConfig

from django_better_repr.signals import auto_configure


class DjangoBetterReprAutoConfigure(AppConfig):
    default = True
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'django_better_repr'

    def ready(self):
        auto_configure()



from django.db import models

from django_better_repr import better_repr

# Create your models here.


class JustForFKs(models.Model):
    ...


class JustForM2Ms(models.Model):
    ...


@better_repr
class FourOrLessFields(models.Model):
    one = models.CharField(max_length=16)
    two = models.CharField(max_length=32, default='Default', null=True)
    three = models.IntegerField(null=True)
    four = models.ForeignKey(to=JustForFKs, on_delete=models.CASCADE)


@better_repr
class FourOrMoreFields(models.Model):
    one = models.CharField(max_length=16)
    two = models.CharField(max_length=32, default='Default', null=True)
    three = models.IntegerField(null=True)
    four = models.ForeignKey(to=JustForFKs, on_delete=models.CASCADE)
    five = models.ManyToManyField(to=JustForM2Ms)
    six = models.FloatField()


class WithoutDecorator(models.Model):
    one = models.CharField(max_length=16)
    two = models.CharField(max_length=32, default='Default', null=True)



from functools import lru_cache
from unittest.mock import sentinel

from django.conf import settings

ALL_APPS = sentinel.ALL_APPS


@lru_cache(maxsize=1)
def config():
    settings_attr = getattr(settings, 'BETTER_REPR_CONFIG', {})

    return {
        # Which models to automatically to include
        'AUTO_CONFIGURE_INCLUDE_MODELS': settings_attr.get('AUTO_CONFIGURE_INCLUDE_MODELS', ALL_APPS),
        # Which models to exclude from automatic inclusion
        'AUTO_CONFIGURE_EXCLUDE_MODELS': settings_attr.get('AUTO_CONFIGURE_EXCLUDE_MODELS', []),
        # Enable multiline REPRs
        'ENABLE_MULTILINE_REPRS': settings_attr.get('ENABLE_MULTILINE_REPRS', True),
        # Should deferred fields be included (1 query per deferred field)
        'EXCLUDE_DEFERRED_FIELDS': settings_attr.get('EXCLUDE_DEFERRED_FIELDS', True),
        # What whitespace to use for multiline REPRs, this is so we don't need to debate spaces vs tabs
        'MULTILINE_WHITESPACE': settings_attr.get('MULTILINE_WHITESPACE', '\t'),
        # How many parts (example: id=1) before splitting into a new line
        'SINGLE_LINE_PARTS_LIMIT': settings_attr.get('SINGLE_LINE_PARTS_LIMIT', 4),
    }



"""
The code in this module borrows heavily from https://github.com/dan-passaro/django-auto-repr/
which requires the MIT license to be included:

MIT License

Copyright (c) 2017 Daniel Passaro

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from functools import partial

from django.db import models
from django.utils.datastructures import ImmutableList

from django_better_repr.config import config
from django_better_repr.exceptions import BetterReprException


def _repr_format_field(self, field):
    if isinstance(field, models.ManyToManyField):
        return

    if isinstance(field, models.ForeignKey):
        field_name = f"{field.name}_id"
    else:
        field_name = field.name

    field_value = getattr(self, field_name)

    default = field.default
    if field_value == default:
        return ''
    elif default is models.NOT_PROVIDED:
        if isinstance(field, models.CharField) and field_value == '':
            return ''
        if field_value is None:
            return ''
    return f"{field_name}={repr(field_value)}"


class BetterRepr:
    def __repr__(self):
        cls = type(self)
        try:
            if config()['EXCLUDE_DEFERRED_FIELDS']:
                all_fields: ImmutableList = cls._meta.fields
                deferred_fields: set = self.get_deferred_fields()
                fields = [f for f in all_fields if f.name not in deferred_fields]
            else:
                if deferred_fields := self.get_deferred_fields():
                    self.refresh_from_db(fields=deferred_fields)
                all_fields = cls._meta.fields
                fields = [f for f in all_fields]

        except AttributeError as e:
            raise BetterReprException(f"{cls} does not appear to be a django model!") from e

        parts = list(
            filter(
                None,
                map(
                    partial(
                        _repr_format_field,
                        self,
                    ),
                    fields,
                ),
            ),
        )
        if len(parts) > config()['SINGLE_LINE_PARTS_LIMIT'] and config()['ENABLE_MULTILINE_REPRS']:
            attrs = ',\n\t'.join(parts)
            return f"{cls.__name__}(\n{config()['MULTILINE_WHITESPACE']}{attrs},\n)"
        else:
            attrs = ', '.join(parts)
            return f'{cls.__name__}({attrs})'



# Generated by Django 4.2.2 on 2023-06-30 22:41

import django.db.models.deletion
from django.db import migrations, models

import django_better_repr.bases


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='JustForFKs',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            bases=(django_better_repr.bases.BetterRepr, models.Model),
        ),
        migrations.CreateModel(
            name='JustForM2Ms',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            bases=(django_better_repr.bases.BetterRepr, models.Model),
        ),
        migrations.CreateModel(
            name='WithoutDecorator',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('one', models.CharField(max_length=16)),
                ('two', models.CharField(default='Default', max_length=32, null=True)),
            ],
            bases=(django_better_repr.bases.BetterRepr, models.Model),
        ),
        migrations.CreateModel(
            name='FourOrMoreFields',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('one', models.CharField(max_length=16)),
                ('two', models.CharField(default='Default', max_length=32, null=True)),
                ('three', models.IntegerField(null=True)),
                ('six', models.FloatField()),
                ('five', models.ManyToManyField(to='myapp.justform2ms')),
                ('four', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='myapp.justforfks')),
            ],
            bases=(django_better_repr.bases.BetterRepr, models.Model),
        ),
        migrations.CreateModel(
            name='FourOrLessFields',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('one', models.CharField(max_length=16)),
                ('two', models.CharField(default='Default', max_length=32, null=True)),
                ('three', models.IntegerField(null=True)),
                ('four', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='myapp.justforfks')),
            ],
            bases=(django_better_repr.bases.BetterRepr, models.Model),
        ),
    ]



import sys

from django_better_repr.bases import BetterRepr

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import PackageNotFoundError, version  # pragma: no cover
else:
    from importlib_metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = "django-better-repr"
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError


def better_repr(klass: type = None, /, ):
    def decorator(kls: type):
        if BetterRepr not in kls.__bases__:
            kls.__bases__ = (BetterRepr, *kls.__bases__)
        return kls

    if klass:
        return decorator(klass)
    else:
        return decorator



import pytest

from tests.example_project.myapp.models import (
    FourOrLessFields,
    FourOrMoreFields,
    JustForFKs,
    WithoutDecorator,
)


def test_with_four_or_less_fields():
    instance = FourOrLessFields()
    actual = repr(instance)
    expected = "FourOrLessFields()"

    assert actual == expected


@pytest.mark.parametrize(
    'data, expected',
    [
        ({}, 'FourOrLessFields(id=1, four_id=1)'),
        ({'one': 'Some Value'}, "FourOrLessFields(id=1, one='Some Value', four_id=1)"),
        ({'one': '', 'two': None, 'three': None}, 'FourOrLessFields(id=1, two=None, four_id=1)'),
        ({'one': '', 'two': '', 'three': '77'}, "FourOrLessFields(id=1, two='', three='77', four_id=1)"),
        ({'one': '', 'two': 'Something', 'three': '77'}, "FourOrLessFields(id=1, two='Something', three='77', four_id=1)"),
    ]
)
@pytest.mark.django_db
def test_with_four_or_less_fields_saved(data, expected):
    instance = FourOrLessFields.objects.create(
        **data,
        four=JustForFKs.objects.create()
    )
    actual = repr(instance)
    assert actual == expected


@pytest.mark.parametrize(
    'data, expected',
    [
        ({}, 'FourOrLessFields(id=1, four_id=1)'),
        ({'one': 'Some Value'}, "FourOrLessFields(id=1, one='Some Value', four_id=1)"),
        ({'one': '', 'two': None, 'three': None}, 'FourOrLessFields(id=1, two=None, four_id=1)'),
        ({'one': '', 'two': '', 'three': '77'}, "FourOrLessFields(id=1, two='', three=77, four_id=1)"),
        ({'one': '', 'two': 'Something', 'three': '77'}, "FourOrLessFields(id=1, two='Something', three=77, four_id=1)"),
    ]
)
@pytest.mark.django_db
def test_with_four_or_less_fields_saved_and_refreshed(data, expected):
    instance = FourOrLessFields.objects.create(
        **data,
        four=JustForFKs.objects.create()
    )
    instance.refresh_from_db()
    actual = repr(instance)
    assert actual == expected


@pytest.mark.django_db
def test_with_four_or_more_fields():
    instance = FourOrMoreFields(
        one='one',
        two='two',
        three='three',
        four=JustForFKs.objects.create(),
        six=99.9,
    )
    actual = repr(instance)
    expected = "FourOrMoreFields(\n\tone='one',\n\ttwo='two',\n\tthree='three',\n\tfour_id=1,\n\tsix=99.9,\n)"

    assert actual == expected


@pytest.mark.django_db
def test_auto_configure_works():
    instance = WithoutDecorator.objects.create(
        one='one',
        two='two',
    )
    actual = repr(instance)
    expected = "WithoutDecorator(id=1, one='one', two='two')"
    assert actual == expected


@pytest.mark.django_db
def test_deferred_fields_are_not_loaded_by_default():
    WithoutDecorator.objects.create(
        one='one',
        two='two',
    )
    instance = WithoutDecorator.objects.all().only('id', 'one').first()

    actual = repr(instance)
    expected = "WithoutDecorator(id=1, one='one')"
    assert actual == expected


@pytest.mark.django_db
def test_deferred_fields_are_loaded(better_repr_config):
    better_repr_config['EXCLUDE_DEFERRED_FIELDS'] = False
    WithoutDecorator.objects.create(
        one='one',
        two='two',
    )
    instance = WithoutDecorator.objects.all().only('id', 'one').first()

    actual = repr(instance)
    expected = "WithoutDecorator(id=1, one='one', two='two')"
    assert actual == expected

    better_repr_config['EXCLUDE_DEFERRED_FIELDS'] = True



from django.apps import apps
from django.core.signals import setting_changed
from django.dispatch import receiver

from django_better_repr import better_repr
from django_better_repr.config import ALL_APPS, config


def auto_configure():
    include = config()['AUTO_CONFIGURE_INCLUDE_MODELS']
    if include is ALL_APPS:
        include = apps.get_models()

    exclude = {x for x in config()['AUTO_CONFIGURE_EXCLUDE_MODELS']}
    models = {x for x in include}.difference(exclude)

    for model in models:
        better_repr()(model)


@receiver(setting_changed)
def settings_changed_callback(sender, setting, value, enter, **kwargs):
    if setting == 'BETTER_REPR_CONFIG':
        config.cache_clear()
